#include "toposens_markers/plot.h"
#include <toposens_driver/sensor.h>


namespace toposens_markers
{
/** A dynamic reconfigure server is set up to change marker scale
 *  and lifetime during runtime.
 */
Plot::Plot(ros::NodeHandle nh, ros::NodeHandle private_nh)
{
	private_nh.param<std::string>("frame_id", _frame_id, "toposens");		//Frame for tf
  _rviz.reset(new rviz_visual_tools::RvizVisualTools(_frame_id, "/" + kMarkersTopic));
  _rviz->enableBatchPublishing();

  // Set up dynamic reconfigure to change marker settings
  _srv = std::make_shared<Cfg>(private_nh);
  Cfg::CallbackType f = boost::bind(&Plot::_reconfig, this, _1, _2);
  _srv->setCallback(f);

	// Subscribe to topic with TS scans
	_scans_sub = nh.subscribe(toposens_driver::kScansTopic, 100, &Plot::_sanitize, this);
  _sensingRange = 0;
}


/** Current implementation defines 2 marker parameters: scale and
 *  lifetime. On each trigger, the latest config data structure is
 *  copied locally and the user-defined global scale is updated for
 *  rviz rendering. No separate updating is done for the lifetime
 *  parameter as it is always retrieved directly from the cfg object
 *  whenever its value is needed in the code.
 */
void Plot::_reconfig(TsMarkersConfig &cfg, uint32_t level)
{
  if ((int)level > 2) {
    ROS_INFO("Update skipped: Parameter not recognized");
    return;
  }

  _lifetime = cfg.lifetime;
  _rviz->setGlobalScale(cfg.scale);
  if (level == -1) ROS_INFO("Marker settings initialized");
  else ROS_INFO("Marker settings updated");
}


/** All incoming scans are stored in a double-ended queue, which
 *  allows for easy iteration, sequential arrangement of data and
 *  efficient chronological cleaning in subsequent runs. On each
 *  trigger, all scans that have expired their lifetime are deleted.
 *  Rviz plotting is thereafter refreshed for this sanitized data.
 */
void Plot::_sanitize(const toposens_msgs::TsScan::ConstPtr& msg)
{
	_scans.push_back(*msg);

  do {
    ros::Time oldest = _scans.front().header.stamp;
    // break loop if oldest timestamp is within lifetime
		if (ros::Time::now() - oldest < ros::Duration(_lifetime)) break;
    _scans.pop_front();
  } while(!_scans.empty());

  Plot::_plot();
}


/** Rviz provides no straightforward way of deleting individual
 *  markers once they have been plotted. Refreshing of display is
 *  achieved by clearing all markers on each run and re-plotting
 *  all data contained in the most recent sanitized copy of incoming
 *  scans.
 *
 *  Sensing range is also updated to keep track of the furthest point
 *  detected by the sensor. This is used for color-coding the markers.
 *
 *  RVT (Rviz Visual Tools) wrappers are used for efficient Rviz
 *  plotting. Since batch publishing is enabled, RVT collects all
 *  markers to be visualized in a given update and publishes them
 *  to Rviz in one go when trigger() is called.
 */
void Plot::_plot(void) {
  _rviz->deleteAllMarkers();

  for (auto it1 = _scans.begin(); it1 != _scans.end(); ++it1) {
		std::vector<toposens_msgs::TsPoint> points = it1->points;

    for (auto it2 = points.begin(); it2 != points.end(); ++it2) {
			toposens_msgs::TsPoint pt = *it2;
      if (pt.location.x > _sensingRange) _sensingRange = pt.location.x;

      Eigen::Vector3d location = _rviz->convertPoint32(pt.location);
      std_msgs::ColorRGBA color = _rviz->getColorScale(pt.location.x/_sensingRange);
      geometry_msgs::Vector3 scale = _rviz->getScale(_baseScale, pt.intensity);

      if (scale.x > 0) _rviz->publishSphere(location, color, scale, kMarkersNs);
    }
  }
  _rviz->trigger();
}


} // namespace toposens_markers
